-------------------------------------------------------------------------------
-- Title      : Crossbar Switch Package
-- Project    : AXI-3 Crossbar Switch
-------------------------------------------------------------------------------
-- File       : axi3-interconnect_pkg.vhd
-- Author     : Mario Hüttel <mario.huettel@gmx.net>
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: 
-------------------------------------------------------------------------------
-- Copyright (c) 2016
--
-- This file is part of AXI-3 Crossbar Switch.
--
-- AXI-3 Crossabr Switch is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation,  version 2 of the License.
--
-- This code is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this code.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package axi3intercon_pkg is
	-- Slave bits have to be more. SLAVE_BITS = MASTER_BITS + log_2(MASTER_COUNT)
	-- Master ID is padded onto ID to identify the transfer internally in the crossbar
	-- This  is defined in AXI-3 and -4 Standard.
	constant RID_MASTER_BITS : natural := 12;
	constant RID_SLAVE_BITS  : natural := 14;
	constant WID_MASTER_BITS : natural := 12;
	constant WID_SLAVE_BITS  : natural := 14;

	constant DATA_BITS    : natural := 32;
	constant DATA_STROBES : natural := (DATA_BITS / 8);
	constant ADDRESS_BITS : natural := 32;

	constant USER_BITS : natural := 4;

	constant MASTER_COUNT : natural := 2;
	constant SLAVE_COUNT  : natural := 2;

	-- Constant definitions for signals

	--Definitions for burst size
	constant AXI_SIZE_1   : std_logic_vector(2 downto 0) := (others => '0');
	constant AXI_SIZE_2   : std_logic_vector(2 downto 0) := "001";
	constant AXI_SIZE_4   : std_logic_vector(2 downto 0) := "010";
	constant AXI_SIZE_8   : std_logic_vector(2 downto 0) := "011";
	constant AXI_SIZE_16  : std_logic_vector(2 downto 0) := "100";
	constant AXI_SIZE_32  : std_logic_vector(2 downto 0) := "101";
	constant AXI_SIZE_64  : std_logic_vector(2 downto 0) := "110";
	constant AXI_SIZE_128 : std_logic_vector(2 downto 0) := "111";

	-- Definitions for response vector
	constant AXI_RESP_OKAY   : std_logic_vector(1 downto 0) := "00";
	constant AXI_RESP_EXOKAY : std_logic_vector(1 downto 0) := "01";
	constant AXI_RESP_SLVERR : std_logic_vector(1 downto 0) := "10";
	constant AXI_RESP_DECERR : std_logic_vector(1 downto 0) := "11";

	-- Definitions for lock locked transactions (only AXI 3 supports locked access mode)
	constant AXI_LOCK_NORMAL    : std_logic_vector(1 downto 0) := "00";
	constant AXI_RESP_EXCLUSIVE : std_logic_vector(1 downto 0) := "01";
	constant AXI_RESP_LOCKED    : std_logic_vector(1 downto 0) := "10";

	-- type declarations for AW channel
	type master_aw_out_t is record
		awid    : std_logic_vector(WID_MASTER_BITS - 1 downto 0); -- Write transaction ID
		awaddr  : std_logic_vector(DATA_BITS - 1 downto 0); -- Write address
		awlen   : std_logic_vector(7 downto 0); -- Burst length - 1;
		awsize  : std_logic_vector(2 downto 0); -- Maximum size of the beat
		awburst : std_logic_vector(1 downto 0); -- Burst type
		awlock  : std_logic_vector(1 downto 0); -- Lock info
		awcache : std_logic_vector(3 downto 0); -- caching info
		awprot  : std_logic_vector(2 downto 0); -- protection info
		awvalid : std_logic;            -- Data valid
	end record master_aw_out_t;

	type master_aw_in_t is record
		awready : std_logic;
	end record master_aw_in_t;

	type slave_aw_in_t is record
		awid    : std_logic_vector(WID_SLAVE_BITS - 1 downto 0); -- Write transaction ID
		awaddr  : std_logic_vector(DATA_BITS - 1 downto 0); -- Write address
		awlen   : std_logic_vector(7 downto 0); -- Burst length - 1;
		awsize  : std_logic_vector(2 downto 0); -- Maximum size of the beat
		awburst : std_logic_vector(1 downto 0); -- Burst type
		awlock  : std_logic_vector(1 downto 0); -- Lock info
		awcache : std_logic_vector(3 downto 0); -- caching info
		awprot  : std_logic_vector(2 downto 0); -- protection info
		awvalid : std_logic;            -- Data valid
	end record slave_aw_in_t;

	subtype slave_aw_out_t is master_aw_in_t;

	-- type decalatations for AR cahnnel
	type master_ar_out_t is record
		arid    : std_logic_vector(WID_MASTER_BITS - 1 downto 0); -- Write transaction ID
		araddr  : std_logic_vector(DATA_BITS - 1 downto 0); -- Write start address
		arlen   : std_logic_vector(7 downto 0); -- Burst length - 1;
		arsize  : std_logic_vector(2 downto 0); -- Maximum size of the beat
		arburst : std_logic_vector(1 downto 0); -- Burst type
		arlock  : std_logic_vector(1 downto 0); -- Lock info
		arcache : std_logic_vector(3 downto 0); -- caching info
		arprot  : std_logic_vector(2 downto 0); -- protection info
		arvalid : std_logic;            -- Data valid
	end record master_ar_out_t;

	type master_ar_in_t is record
		arready : std_logic;
	end record master_ar_in_t;

	type slave_ar_in_t is record
		arid    : std_logic_vector(WID_SLAVE_BITS - 1 downto 0); -- Read transaction ID
		araddr  : std_logic_vector(DATA_BITS - 1 downto 0); -- Read start address
		arlen   : std_logic_vector(7 downto 0); -- Burst length - 1;
		arsize  : std_logic_vector(2 downto 0); -- Maximum size of the beat
		arburst : std_logic_vector(1 downto 0); -- Burst type
		arlock  : std_logic_vector(1 downto 0); -- Lock info
		arcache : std_logic_vector(3 downto 0); -- caching info
		arprot  : std_logic_vector(2 downto 0); -- protection info
		arvalid : std_logic;            -- Data valid
	end record slave_ar_in_t;

	subtype slave_ar_out_t is master_ar_in_t;

	-- type decalarations for W channel
	type master_w_out_t is record
		wid    : std_logic_vector(WID_MASTER_BITS - 1 downto 0);
		wdata  : std_logic_vector(DATA_BITS - 1 downto 0);
		wstrb  : std_logic_vector(DATA_STROBES - 1 downto 0);
		wlast  : std_logic;
		wuser  : std_logic_vector(USER_BITS - 1 downto 0); -- user defined
		wvalid : std_logic;
	end record master_w_out_t;

	type slave_w_in_t is record
		wid    : std_logic_vector(WID_SLAVE_BITS - 1 downto 0);
		wdata  : std_logic_vector(DATA_BITS - 1 downto 0);
		wstrb  : std_logic_vector(DATA_STROBES - 1 downto 0);
		wlast  : std_logic;
		wuser  : std_logic_vector(USER_BITS - 1 downto 0); -- user defined
		wvalid : std_logic;
	end record slave_w_in_t;

	type master_w_in_t is record
		wready : std_logic;
	end record master_w_in_t;

	subtype slave_w_out_t is master_w_in_t;

	-- Type declarations for R channel
	type master_r_in_t is record
		rid    : std_logic_vector(RID_MASTER_BITS - 1 downto 0);
		rdata  : std_logic_vector(DATA_BITS - 1 downto 0);
		rresp  : std_logic_vector(1 downto 0);
		rlast  : std_logic;
		ruser  : std_logic_vector(USER_BITS - 1 downto 0);
		rvalid : std_logic;
	end record master_r_in_t;

	type slave_r_out_t is record
		rid    : std_logic_vector(RID_SLAVE_BITS - 1 downto 0);
		rdata  : std_logic_vector(DATA_BITS - 1 downto 0);
		rresp  : std_logic_vector(1 downto 0);
		rlast  : std_logic;
		ruser  : std_logic_vector(USER_BITS - 1 downto 0);
		rvalid : std_logic;
	end record slave_r_out_t;

	type master_r_out_t is record
		rready : std_logic;
	end record master_r_out_t;

	subtype slave_r_in_t is master_r_out_t;

	-- Type declarations for B channel
	type master_b_in_t is record
		bid    : std_logic_vector(WID_MASTER_BITS - 1 downto 0);
		bresp  : std_logic_vector(1 downto 0);
		buser  : std_logic_vector(USER_BITS - 1 downto 0);
		bvalid : std_logic;
	end record master_b_in_t;

	type slave_b_out_t is record
		bid    : std_logic_vector(WID_SLAVE_BITS - 1 downto 0);
		bresp  : std_logic_vector(1 downto 0);
		buser  : std_logic_vector(USER_BITS - 1 downto 0);
		bvalid : std_logic;
	end record slave_b_out_t;

	type master_b_out_t is record
		bready : std_logic;
	end record master_b_out_t;

	subtype slave_b_in_t is master_b_out_t;

	-- Combined definitions

	type axi_master_in_t is record
		aw : master_aw_in_t;
		ar : master_ar_in_t;
		w  : master_w_in_t;
		r  : master_r_in_t;
		b  : master_b_in_t;
	end record axi_master_in_t;

	type axi_master_out_t is record
		aw : master_aw_out_t;
		ar : master_ar_out_t;
		w  : master_w_out_t;
		r  : master_r_out_t;
		b  : master_b_out_t;
	end record axi_master_out_t;

	type axi_slave_in_t is record
		aw : slave_aw_in_t;
		ar : slave_ar_in_t;
		w  : slave_w_in_t;
		r  : slave_r_in_t;
		b  : slave_b_in_t;
	end record axi_slave_in_t;

	type axi_slave_out_t is record
		aw : slave_aw_out_t;
		ar : slave_ar_out_t;
		w  : slave_w_out_t;
		r  : slave_r_out_t;
		b  : slave_b_out_t;
	end record axi_slave_out_t;

	-- Connection array definitions
	type axi_masters_in_t is array (natural range <>) of axi_master_in_t;
	type axi_masters_out_t is array (natural range <>) of axi_master_out_t;
	type axi_slaves_in_t is array (natural range <>) of axi_slave_in_t;
	type axi_slaves_out_t is array (natural range <>) of axi_slave_out_t;

	-- Address translation mapping
	type axi_slave_addresses_t is array (natural range <>) of std_logic_vector(ADDRESS_BITS - 1 downto 0);

	function calculate_slave(address       : in std_logic_vector(ADDRESS_BITS - 1 downto 0);
			                 address_array : in axi_slave_addresses_t(0 to SLAVE_COUNT - 1);
			                 mask_array    : in axi_slave_addresses_t(0 to SLAVE_COUNT - 1)) return integer;

end package axi3intercon_pkg;

package body axi3intercon_pkg is
	function calculate_slave(address       : in std_logic_vector(ADDRESS_BITS - 1 downto 0);
			                 address_array : in axi_slave_addresses_t(0 to SLAVE_COUNT - 1);
			                 mask_array    : in axi_slave_addresses_t(0 to SLAVE_COUNT - 1)) return integer is
		variable slave_idx : integer range 0 to SLAVE_COUNT := SLAVE_COUNT;
	begin
		slave_idx := SLAVE_COUNT;
		for i in 0 to SLAVE_COUNT - 1 loop
			if (address and mask_array(i)) = address_array(i) then
				slave_idx := i;
			end if;
		end loop;
		return slave_idx;
	end function calculate_slave;

end package body axi3intercon_pkg;
